#!/usr/bin/perl

use	XML::Parser;

BEGIN {
	# Set version
	my $cvsId = '$Id: dia2postgres.pl,v 1.16 2002/10/05 13:06:06 uid5000 Exp $';
	$cvsId =~ s/^\$Id: //;
	$cvsId =~ s/\$$//;
	$main::VERSION	= $cvsId;
}

# Activate debugging by setting those variables to 1
$DEBUG		= 0;
$DEBUGXML	= 0;

# Generic parsing
$prefix		= '';
$xmlLevel	= 0;
@allObjects	= ();
@customCode	= ();
@enumerations	= ();

# Group handling
$inGroup	= 0;
$subGroups	= 0;
$noGroup	= 0;
@groups		= ();
%groupsOrdered	= ();
$inLPackage	= 0;

# Relations and inheritance
$currentLine	= '';
%lines		= ();
$inLine		= 0;
$inLineType	= 0;
$inLConnections	= 0;

# Class handling
%classes	= ();
%classesName	= ();
$currentClass	= '';
$currentField	= '';
$currentIndex	= '';
$inClass	= 0;
$inClassName	= 0;
$inClassSte	= 0;
$inClassAttrs	= 0;
$inClassField	= 0;
$inClassOps	= 0;
$inClassIndex	= 0;

# Table field handling
$inCFName	= 0;
$inCFType	= 0;
$inCFOptions	= 0;
$inCFValue	= 0;

# Table index handling
$inCIName	= 0;
$inCIType	= 0;
$inCIUniqueness	= 0;
$inCIParams	= 0;
$inCIField	= 0;
$inCIFieldName	= 0;

# Text handling
$inStdText	= 0;
$totalText	= '';


###                        ###
#### XML PARSER FUNCTIONS ####
###                        ###

###
## Called if parser enters a tag
#
sub	StartTag
{
	my $p = shift;				# parser context
	$ctag = shift;				# name of this tag element
	%attr = %_;				# hash with attributes

	$DEBUGXML && print STDERR (" "x($xmlLevel += 1)) . "> $ctag\n";

	# Handle diagram start for dia >= 0.83.0
	if	($ctag eq $prefix.'dia:diagram')
	{
		$prefix = 'dia:';
	}
	# Handle groups
	elsif	($ctag eq $prefix.'group')
	{
		if	($inGroup)
		{
			$subGroups ++;
			$DEBUG && print STDERR "Entering subgroup level $subGroup for group $noGroup\n";
		}
		else
		{
			$DEBUG && print STDERR "Entering top-level group $noGroup\n";
			$inGroup = 1;
			$groups[$noGroup] = {
				to		=> 'group',
				type		=> '',		# Either 'box' or 'pack'
				topLevelText	=> '',
				order		=> 0,
				classes		=> []
			};
			push @allObjects, $groups[$noGroup];
		}
	}
	# Handle standard text
	elsif	($inStdText == 1 && $ctag eq $prefix.'attribute' && $attr{name} eq 'text')
	{
		$DEBUG && print STDERR " > found 'text' attribute\n";
		$inStdText = 2;
	}
	elsif	($inStdText == 2 && $ctag eq $prefix.'string')
	{
		$DEBUG && print STDERR "  > found string\n";
		$inStdText = 3;
		$totalText = '';
	}
	elsif	($inStdText)
	{
		return;
	}
	# Package name lookup
	elsif	($inLPackage == 1 && $ctag eq $prefix.'attribute' && $attr{name} eq 'name')
	{
		$inLPackage++;
	}
	elsif	($inLPackage == 2 && $ctag eq $prefix.'string')
	{
		$inLPackage++;
		$totalText = '';
	}
	# Class name lookup
	elsif	($inClassName == 1 && $ctag eq $prefix.'string')
	{
		$inClassName++;
		$totalText = '';
	}
	# Class stereotype lookup
	elsif	($inClassSte == 1 && $ctag eq $prefix.'string')
	{
		$inClassSte++;
		$totalText = '';
	}
	# Table field name lookup
	elsif	($inCFName == 1 && $ctag eq $prefix.'string')
	{
		$inCFName++;
		$totalText = '';
	}
	# Table field type lookup
	elsif	($inCFType == 1 && $ctag eq $prefix.'string')
	{
		$inCFType++;
		$totalText = '';
	}
	# Table field options lookup
	elsif	($inCFOptions == 1 && $ctag eq $prefix.'enum')
	{
		if	($attr{val} eq '0')
		{
			$currentField->{mode} = 1;
			$currentField->{not_null} = 1;
		}
		elsif	($attr{val} eq '2')
			{ $currentField->{mode} = 2; }
		$inCFOptions = 0;
	}
	# Table field default value lookup
	elsif	($inCFValue == 1 && $ctag eq $prefix.'string')
	{
		$inCFValue++;
		$totalText = '';
	}
	# Table field 'NOT NULL' option
	elsif	($inCFNotNull == 1 && $ctag eq $prefix.'boolean')
	{
		$inCFNotNull = 0;
		if	($attr{val} eq 'true')
		{
			$currentField->{not_null} = 1;
		}
	}
	# Handle table fields
	elsif	($inClassField && $ctag eq $prefix.'attribute')
	{
		$attrLevel++;
		if	($attr{name} eq 'name')
			{ $inCFName = 1; }
		elsif	($attr{name} eq 'type')
			{ $inCFType = 1; }
		elsif	($attr{name} eq 'visibility')
			{ $inCFOptions = 1; }
		elsif	($attr{name} eq 'value')
			{ $inCFValue = 1; }
		elsif	($attr{name} eq 'class_scope')
			{ $inCFNotNull = 1; }
	}
	# Table index name lookup
	elsif	($inCIName == 1 && $ctag eq $prefix.'string')
	{
		$inCIName++;
		$totalText = '';
	}
	# Table index type lookup
	elsif	($inCIType == 1 && $ctag eq $prefix.'string')
	{
		$inCIType++;
		$totalText = '';
	}
	# Table index uniqueness check
	elsif	($inCIUniqueness == 1 && $ctag eq $prefix.'enum')
	{
		if	($attr{val} eq '1')
		{
			$currentIndex->{unique} = 1;
			$DEBUG && print STDERR " > Current index is an unique index\n";
		}
		$inCIUniqueness = 0;
	}
	# Table index fields
	elsif	($inCIFieldName == 1 && $ctag eq $prefix.'string')
	{
		$inCIFieldName++;
		$totalText = '';
	}
	elsif	($inCIField == 1 && $ctag eq $prefix.'attribute')
	{
		$attrLevel++;
		$inCIFieldName = 1 if $attr{name} eq 'name';
	}
	elsif	($inCIParams == 1 && $ctag eq $prefix.'composite' && $attr{type} eq 'umlparameter')
	{
		$inCIField = 1;
	}
	# Handle table indices
	elsif	($inClassIndex && $ctag eq $prefix.'attribute')
	{
		$attrLevel++;
		if	($attr{name} eq 'name')
			{ $inCIName = 1; }
		elsif	($attr{name} eq 'type')
			{ $inCIType = 1; }
		elsif	($attr{name} eq 'visibility')
			{ $inCIUniqueness = 1; }
		elsif	($attr{name} eq 'parameters')
			{ $inCIParams = 1; }
	}
	# Handle tables
	elsif	($inClass)
	{
		if	($ctag eq $prefix.'attribute' && $attr{name} eq 'name' && !$attrLevel)
		{
			$inClassName = 1;
			$attrLevel++;
		}
		elsif	($ctag eq $prefix.'attribute' && $attr{name} eq 'stereotype' && !$attrLevel)
		{
			$inClassSte = 1;
			$attrLevel++;
		}
		elsif	($ctag eq $prefix.'attribute' && $attr{name} eq 'attributes' && !$attrLevel)
		{
			$DEBUG && print STDERR " > start of attributes list\n";
			$inClassAttrs = 1;
		}
		elsif	($ctag eq $prefix.'composite' && $attr{type} eq 'umlattribute' && $inClassAttrs)
		{
			$DEBUG && print STDERR "  > start of attribute\n";
			$inClassField = 1;
			push (@{$currentClass->{fields}}, ($currentField = {
				to		=> 'field',
				parent		=> $currentClass,
				name		=> '',
				type		=> '',
				mode		=> 0,			# 1 = primary key
									# 2 = foreign key
				not_null	=> 0,
				references	=> '',
				default		=> ''
			}));
			push @allObjects, $currentField;
		}
		elsif	($ctag eq $prefix.'attribute' && $attr{name} eq 'operations' && !$attrLevel)
		{
			$DEBUG && print STDERR " > start of operations list\n";
			$inClassOps = 1;
		}
		elsif	($ctag eq $prefix.'composite' && $attr{type} eq 'umloperation' && $inClassOps)
		{
			$DEBUG && print STDERR "  > start of index\n";
			$inClassIndex = 1;
			push (@{$currentClass->{indices}}, ($currentIndex = {
				to		=> 'index',
				parent		=> $currentClass,
				unique		=> 0,			# [UNIQUE]
				type		=> '',			# USING type
				name		=> '',
				fields		=> []
			}));
			push @allObjects, $currentIndex;
		}
		elsif	($ctag eq $prefix.'attribute')
		{
			$attrLevel++;
		}
	}
	# Handle lines
	elsif	($inLConnections && $ctag eq $prefix.'connection' && $inLine == 1)
	{
		my	$h = {
			class	=> $attr{to},
			dest	=> $attr{connection}
		};
		if	($attr{handle} eq '1')
			{ $currentLine->{to} = $h; }
		elsif	($attr{handle} eq '0')
			{ $currentLine->{from} = $h; }
	}
	elsif	($inLConnections && $ctag eq $prefix.'connection')
	{
		my	$h = {
			class	=> $attr{to},
			dest	=> $attr{connection}
		};
		if	($attr{handle} eq '0')
			{ $currentLine->{to} = $h; }
		elsif	($attr{handle} eq '1')
			{ $currentLine->{from} = $h; }
	}
	elsif	($inLineType && $ctag eq $prefix.'enum')
	{
		if	($attr{val} eq '1')
		{
			$currentLine->{type} = 0;
		}
		elsif	($attr{val} eq '6')
		{
			$currentLine->{type} = 1;
		}
		elsif	($attr{val} eq '7')
		{
			$currentLine->{type} = 2;
		}
		elsif	($attr{val} eq '12')
		{
			$currentLine->{type} = 3;
		}
		elsif	($attr{val} eq '2')
		{
			$currentLine->{type} = 4;
		}
	}
	elsif	($inLine)
	{
		if	($ctag eq $prefix.'attribute' && $attr{name} eq 'end_arrow' && $inLine == 1)
		{
			$inLineType = 1;
		}
		elsif	($ctag eq $prefix.'connections')
		{
			$inLConnections = 1;
		}
	}
	# Handle objects
	elsif	(	$ctag eq $prefix.'object' && $attr{type} eq 'Standard - Text'
		&&	$inGroup == 1 && $subGroups == 0
		)
	{
		$DEBUG && print STDERR "Found standard text in top-level group $noGroup\n";
		$inStdText = 1;
	}
	elsif	(	$ctag eq $prefix.'object' && $attr{type} eq 'Standard - Box'
		&&	($groups[$noGroup]->{type} eq '')
		&&	$inGroup == 1 && $subGroups == 0
		)
	{
		$DEBUG && print STDERR "Found box in top-level group $noGroup\n";
		$groups[$noGroup]->{type} = 'box';
	}
	elsif	(	$ctag eq $prefix.'object' && $attr{type} eq 'UML - LargePackage'
		&&	($groups[$noGroup]->{type} eq '')
		&&	$inGroup == 1 && $subGroups == 0
		)
	{
		$DEBUG && print STDERR "Found UML large package in top-level group $noGroup\n";
		$groups[$noGroup]->{type} = 'pack';
		$inLPackage = 1;
	}
	elsif	(	$ctag eq $prefix.'object' && $attr{type} eq 'UML - Class'
		&&	$inGroup == 1
		)
	{
		$DEBUG && print STDERR "Found UML class in top-level group $noGroup\n";
		push (@{$groups[$noGroup]->{classes}}, ($currentClass = {
			to	=> 'class',
			dia_oid	=> $attr{id},
			parent	=> $groups[$noGroup],
			name	=> '',
			order	=> 0,
			fields	=> [],
			inherits=> [],
			indices	=> []
		}));
		$classes{$attr{id}} = $currentClass;
		push @allObjects, $currentClass;
		$inClass = 1;
		$attrLevel = 0;
	}
	elsif	(	$ctag eq $prefix.'object'
		&&	(	$attr{type} eq 'Standard - Line'
			||	$attr{type} eq 'Standard - ZigZagLine'
			)
		)
	{
		$DEBUG && print STDERR "Found line\n";
		push (@allObjects, ($currentLine = {
			to	=> 'relation',
			dia_oid	=> $attr{id},
			type	=> undef,	# 0 = foreign key
						# 1 = foreign key, ON DELETE CASCADE
						# 2 = foreign key, ON UPDATE CASCADE
						# 3 = foreign key, ON UPDATE CASCADE, ON DELETE CASCADE
						# 4 = inheritance
			from	=> undef,
			to	=> undef
		}));
		$lines{$attr{id}} = $currentLine;
		$inLine = 1;
	}
	elsif	($ctag eq $prefix.'object' && $attr{type} eq 'UML - Generalization')
	{
		$DEBUG && print STDERR "Found UML generalization\n";
		push (@allObjects, ($currentLine = {
			to	=> 'relation',
			dia_oid	=> $attr{id},
			type	=> 4,
			from	=> undef,
			to	=> undef
		}));
		$lines{$attr{id}} = $currentLine;
		$inLine = 2;
	}
}


###
## Called if parser leaves a tag
#
sub	EndTag
{
	
	my $p = shift;
	$ctag = shift;

	# Handle end of groups
	if	($ctag eq $prefix.'group')
	{
		if	($subGroups)
		{
			$DEBUG && print STDERR "Leaving subgroup level $subGroup for group $noGroup\n";
			$subGroups --;
		}
		else
		{
			$DEBUG && print STDERR "Leaving top-level group $noGroup\n";
			$inGroup = 0;
			$noGroup++;
		}
	}
	# Handle end of standard text
	elsif	($inStdText == 3 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		if	($inGroup && !$subGroups && $groups[$noGroup]->{topLevelText} eq '')
		{
			$totalText =~ /^([^\(]+)\(([0-9]+)\)$/;
			my	$name	= $1;
			my	$order	= $2;
			$name = $totalText unless $name;
			$order = 0 unless $order;
			$DEBUG && print STDERR "-> totalText = $totalText ; name = $name ; order = $order\n";

			$groups[$noGroup]->{topLevelText} = $name;			
			$groups[$noGroup]->{order} = $order;
			$groupsOrdered{$order} = [] unless $groupsOrdered{$order};
			push @{$groupsOrdered{$order}}, $groups[$noGroup];
			$DEBUG && print STDERR "   > Got top level text for group $noGroup : $totalText\n";
		}
		$inStdText = 1;
	}
	elsif	($inStdText == 1 && $ctag eq $prefix.'object')
	{
		$inStdText = 0;
	}
	# Handle large package name lookup
	elsif	($inLPackage == 3 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		if	($inGroup && !$subGroups)
		{
			($groups[$noGroup]->{topLevelText} ne '') || ($groups[$noGroup]->{topLevelText} = $totalText);
			$DEBUG && print STDERR " > Got name for large package inside top-level group $noGroup : $totalText\n";
		}
		$inLPackage = 1;
	}
	# Class name lookup
	elsif	($inClassName == 2 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		$currentClass->{name} = $totalText;
		$classesName{$totalText} = $currentClass;
		$DEBUG && print STDERR "Current class name = $totalText\n";
		$inClassName = 0;
	}
	# Class stereotype lookup
	elsif	($inClassSte == 2 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		$currentClass->{order} = $totalText;
		$DEBUG && print STDERR "Current class stereotype = $totalText\n";
		$inClassSte = 0;
	}
	# Table field name lookup
	elsif	($inCFName == 2 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		$currentField->{name} = $totalText;
		$DEBUG && print STDERR "Current field name = $totalText\n";
		$inCFName = 0;
	}
	# Table field type lookup
	elsif	($inCFType == 2 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		$currentField->{type} = $totalText;
		$DEBUG && print STDERR "Current field type = $totalText\n";
		$inCFType = 0;
	}
	# Table field default value lookup
	elsif	($inCFValue == 2 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		$currentField->{default} = $totalText;
		$DEBUG && print STDERR "Current field default = $totalText\n";
		$inCFValue = 0;
	}
	# Table fields
	elsif	($inClassField && $ctag eq $prefix.'composite')
	{
		if	($currentField->{mode} == 1)
		{
			$currentClass->{pkey} = ($currentClass->{pkey} ? $currentClass->{pkey} . "," : "") . $currentField->{name};
		}
		$inClassField = 0;
		$DEBUG && print STDERR "  > end of attribute\n";
	}
	elsif	(!$inClassField && $inClassAttrs && $ctag eq $prefix.'attribute')
	{
		$DEBUG && print STDERR " > end of attributes list\n";
		$inClassAttrs = 0;
	}
	# Table index name lookup
	elsif	($inCIName == 2 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		$currentIndex->{name} = $totalText;
		$DEBUG && print STDERR "Current index name = $totalText\n";
		$inCIName = 0;
	}
	# Table index type lookup
	elsif	($inCIType == 2 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		$currentIndex->{type} = $totalText;
		$DEBUG && print STDERR "Current index type = $totalText\n";
		$inCIType = 0;
	}
	# Table index field name lookup
	elsif	($inCIFieldName == 2 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		push @{$currentIndex->{fields}}, $totalText;
		$DEBUG && print STDERR "New index field = $totalText\n";
		$inCIFieldName = 0;
	}
	# Table index fields list
	elsif	($inCIField && $ctag eq $prefix.'composite')
	{
		$inCIField = $inCIFieldName = 0;
	}
	elsif	(!$inCIField && $inCIParams && $ctag eq $prefix.'attribute')
	{
		$inCIParams = 0;
	}
	# Table indices
	elsif	($inClassIndex && $ctag eq $prefix.'composite')
	{
		$inClassIndex = 0;
		if	(!$currentIndex->{name})
		{
			$currentIndex->{name} = "i_" . $currentIndex->{type} . "_";
			$currentIndex->{name} .= $currentIndex->{parent}->{name} . "_";
			$currentIndex->{name} .= join('_', @{$currentIndex->{fields}});
			$DEBUG && print STDERR "New index name = " . $currentIndex->{name} . "\n";
		}
		$DEBUG && print STDERR "  > end of index\n";
	}
	elsif	(!$inClassIndex && $inClassOps && $ctag eq $prefix.'attribute')
	{
		$DEBUG && print STDERR " > end of operations list\n";
		$inClassOps = 0;
	}
	elsif	($inClass && $attrLevel && $ctag eq $prefix.'attribute')
	{
		$attrLevel --;
	}
	# Relation type
	elsif	($inLineType && $ctag eq $prefix.'attribute')
	{
		$inLineType = 0;
	}
	# Handle end of line connections
	elsif	($inLConnections && $ctag eq $prefix.'connections')
	{
		$inLConnections = 0;
	}
	# Handle end of objects
	elsif	($ctag eq $prefix.'object')
	{
		if	($inClass)
		{
			$inClass = 0;
		}
		elsif	($inLine)
		{
			$inLine = 0;
		}
		elsif	($inLPackage)
		{
			$inLPackage = 0;
		}
	}

	$DEBUGXML && print STDERR (" "x($xmlLevel --)) . "> $ctag\n";
}


###
## Called for text between any tags
#
sub	Text
{
	my	$text	= $_;

	if	($totalText eq '')
	{
		$totalText = $text;
		$totalText =~ s/^\#//;
	}
	else
		{ $totalText .= $text; }
}



###                                     ###
#### INPUT VALIDATION AND MODIFICATION ####
###                                     ###

###
## This function creates a valid PostgreSQL identifier from a string
#
sub	strToPgId
{
	my	$string	= shift;

	return $string if $string =~ /^\".*\"$/;

	return "\"$string\"" if (length $string < 32);

	$string =~ s/\_+/_/g;
	return "\"$string\"" if (length $string < 32);

	my	@list	= split /_/, $string;
	my	$old	= $string;
	my	$lsz	= @list;
	my	$rest	= 32 - @list;
	my	$max;

	$string = "";
	foreach my $it (@list)
	{
		$max = $rest / $lsz;
		$max =~ s/\.(.*)$//;
		$max += 1 if ($1 && $1 > 5 && $max < $rest);
		$string .= '_' if $string;

		if	(length $it > $max)
		{
			$string .= substr $it, 0, $max;
			$rest -= $max;
		}
		else
		{
			$string .= $it;
			$rest -= length $string;
		}

		$lsz--;
	}

	print STDERR "WARNING: name \"$old\" changed to \"$string\"\n";
	return "\"$string\"";
}


###
## Validate and modify the whole structure
#
sub	checkAndModifyInput
{
	foreach my $o (@allObjects)
	{
		next unless $o;

		# Handle local fields
		if	(	$o->{to} eq 'field'
			&&	$o->{parent} && $o->{parent}->{parent}
			&&	$o->{parent}->{parent}->{type} eq 'box'
			)
		{
			# Handle enumerations
			if	($o->{type} =~ /^\{ .* \}$/ && !$o->{references})
			{
				my	$t = $o->{type};
				my	@n;
				my	$dsc = {};

				$t =~ s/\{ (.*) \}$/$1/;
				$t =~ s/\s+//g;
				@n = split /,/, $t;
				$t = $o->{parent}->{name} . "." . $o->{name};
				$dsc->{fieldname} = $t;
				$t =~ s/\./_/g;
				$dsc->{compname} = "e_$t";
				$dsc->{tablename} = strToPgId("e_$t");
				$dsc->{contents} = [@n];
				push @enumerations, $dsc;

				$o->{type} = "INT";
				$o->{references} = $dsc->{tablename} . ".id";
				$o->{refCascade} = 3;
			}
		}
		# Sort classes inside groups
		elsif	($o->{to} eq 'group' && $o->{type} eq 'box')
		{
			my $nc = [];
			@$nc = sort { $a->{order} <=> $b->{order} } @{$o->{classes}};
			$o->{classes} = $nc;
		}
	}

	my	@tmp	= grep { $_->{type} == 4 } values %lines;
	foreach my $ir (@tmp)
	{
		my	$target = $classes{$ir->{from}->{class}};
		next unless defined $target;
		my	$super = $classes{$ir->{to}->{class}};
		next unless defined $super;
		push @{$target->{inherits}}, $super;
	}

	@tmp = grep { defined($_->{type}) && ($_->{type} != 4) } values %lines;
	foreach my $rel (@tmp)
	{
		next if $rel->{from}->{dest} < 8;

		my	$target	= $classes{$rel->{from}->{class}};
		next unless defined $target;
		my	$targetFieldNb = $rel->{from}->{dest} - 8;
		$targetFieldNb -= $targetFieldNb % 2;
		$targetFieldNb /= 2;
		next unless $targetFieldNb < @{$target->{fields}};
		my	$targetField = $target->{fields}->[$targetFieldNb];
		next if (	$targetField->{references}
			||	!$targetField->{mode}
			);

		my	$ref	= $classes{$rel->{to}->{class}};
		next unless defined $ref;
		my	$refField = '';

		if	($rel->{to}->{dest} < 8)
		{
			next unless	(	@{$ref->{inherits}}
					||	($ref->{pkey} && $ref->{pkey} !~ /,/)
					);
			if	($ref->{pkey} && $ref->{pkey} !~ /,/)
			{
				$refField = (grep { $_->{name} eq $ref->{pkey} } @{$ref->{fields}})[0];
			}
			elsif	(@{$ref->{inherits}})
			{
				# WARNING : this will loop forever with broken, recurrent inheritance
				my ($tmp, @checklist) = (0);
				push @checklist, @{$ref->{inherits}};
				while (!$refField and $tmp = shift @checklist)
				{
					if	($tmp->{pkey} && $tmp->{pkey} !~ /,/)
						{ $refField = (grep { $_->{name} eq $tmp->{pkey} } @{$tmp->{fields}})[0]; }
					elsif	(@{$tmp->{inherits}})
						{ push @checklist, @{$tmp->{inherits}}; }
				}
			}
		}
		else
		{
			my	$refFieldNb = $rel->{to}->{dest} - 8;
			$refFieldNb -= $refFieldNb % 2;
			$refFieldNb /= 2;
			next unless $refFieldNb < @{$ref->{fields}};
			$refField = $ref->{fields}->[$refFieldNb];
			next if ($refField->{mode} != 1);
		}
		next unless $refField;

		$targetField->{references} = $ref->{name} . '.' . $refField->{name};
		$targetField->{refCascade} = $rel->{type};
	}

	$tmpFunc = sub {
		my $class = shift;
		my $level = shift;
		$DEBUG && print STDERR (" " x $level) . "$class->{name} class : finding inheritance\n";
		$DEBUG && print STDERR (" " x $level) . " > $class->{name} : primary key = $class->{pkey}\n";
		foreach my $superClass (@{$class->{inherits}})
		{
			$DEBUG && print STDERR (" " x $level) . "  > inherits $superClass->{name}\n";
			&$tmpFunc($superClass, $level + 1) unless $superClass->{indone};
			$class->{pkey} .= ($class->{pkey} ? "," : "") . $superClass->{pkey}
				if	$superClass->{pkey};
		}
		$DEBUG && print STDERR (" " x $level) . " > $class->{name} : primary key = $class->{pkey}\n";
		$class->{indone} = 1;
	};
	foreach my $cn (values %classes)
	{
		&$tmpFunc($cn, 0) unless $cn->{indone};
	}
	undef $tmpFunc;
}



###                  ###
#### SQL GENERATION ####
###                  ###

###
## Create complete SQL output
#
sub	createSql
{
	if($DEBUG) { print STDERR "\nWriting SQL statements...\n"; }

	&createSQLHeader();
	&createSQLEnums();
	&createSQLTables();
	&copySQLFiles();
	&createSQLFooter();

	if($DEBUG) { print STDERR "Done!\n\n"; }
}


###
## Create SQL header
#
sub	createSQLHeader
{
	$DEBUG && print STDERR " Writing header\n";
	foreach my $g (@groups)
	{
		next unless ($g->{type} eq 'box' && $g->{topLevelText} ne '' && @{$g->{classes}} == 0);

		my	$title = $g->{topLevelText};
		$title =~ s/\n$//;

		my	@lines = split /\n/, $title;
		foreach my $l (@lines)
			{ print "-- $l\n"; }
		print "\n";
		last;
	}

	print "-- Input file  : $infile\n";
	print "-- Output file : $outfile\n" if $outfile;
	print "\n";
	print "-- Defines the following modules : \n";
	foreach my $g (@groups)
	{
		next unless ($g->{type} eq 'box' && $g->{topLevelText} ne '' && @{$g->{classes}} != 0);

		my	$title = $g->{topLevelText};
		$title =~ s/\s+/ /g;

		print "--  $title\n";
	}
	print "\n";

	my	@requiredModules = ();
	foreach my $g (@groups)
	{
		next unless ($g->{type} eq 'pack' && $g->{topLevelText} ne '' && @{$g->{classes}} != 0);

		my	$title = $g->{topLevelText};
		$title =~ s/\s+/ /g;
		push @requiredModules, $title;
	}
	if	(@requiredModules)
	{
		print "-- Requires the following modules :\n--  ";
		print join ("\n--  ", @requiredModules);
		print "\n\n";
	}
	print "BEGIN TRANSACTION;\n\n\n";
}


###
## Create SQL tables for enumerations according to data created by checkAndModifyInput()
#
sub	createSQLEnums
{
	foreach my $enum (@enumerations)
	{
		my	$l = 0;
		foreach my $w1 (@{$enum->{contents}})
		{
			my	$w1l = length($w1);
			($w1l > $l) && ($l = $w1l);
		}

		print "-- Enumeration table for " . $enum->{fieldname} . "\n\n";
		print "CREATE TABLE " . $enum->{tablename} . "(\n\tid\tSERIAL PRIMARY KEY,\n\t";
		print "txt\tVARCHAR(" . ($l+1) . ") NOT NULL\n);\n";
		foreach my $w2 (@{$enum->{contents}})
		{
			print "INSERT INTO ". $enum->{tablename} . "(txt) VALUES('$w2');\n";
		}
		print "\nINSERT INTO d2p_lst_list(d2p_name, dia_name) VALUES ('" . $enum->{tablename};
		print "', '" . $enum->{compname} . "');\n";
		print "\n\n";
	}
}


###
## Create SQL tables for each included module
#
sub	createSQLTables
{
	foreach my $gro (sort keys %groupsOrdered)
	{
		foreach my $g (@{$groupsOrdered{$gro}})
		{
			next unless ($g->{type} eq 'box' && @{$g->{classes}});

			my	$title = $g->{topLevelText};
			$title =~ s/\n/\n-- /g;
			print "-- Tables for $title\n\n";
			foreach my $o (@{$g->{classes}})
			{
				my	$x = 0;
				my	@pkey = ();
				my	@references = ();
				my	$tnam = strToPgId($o->{name});

				print "CREATE TABLE $tnam (";
				foreach my $f (@{$o->{fields}})
				{
					my	$fnam = strToPgId($f->{name});
					$x ? (print ",") : ($x=1);
					print "\n\t$fnam";
					my	$l = 56 - length $fnam;
					if	($l <= 0)
					{
						print " ";
					}
					else
					{
						my	$nt = $l / 8;
						$nt++ if $nt =~ s/\..*$//;
						print "\t" x $nt;
					}
					print $f->{type};

					if	($f->{not_null})
					{
						print " NOT NULL";
					}
					if	($f->{references})
					{
						# FIXME : has to be an option
						#my	($table, $field) = split /\./, $f->{references};
						#print " REFERENCES $table";
						#$field && print "($field)";
						#print " ON DELETE CASCADE";
						my	($table, $field) = map { strToPgId($_); } (split (/\./, $f->{references}));
						my	$tmpStr = "'$tnam', '$fnam', '$table', '$field', ";

						if	($f->{refCascade} & 1)
							{ $tmpStr .= "true"; }
						else
							{ $tmpStr .= "false"; }
						$tmpStr .= ", ";
						if	($f->{refCascade} & 2)
							{ $tmpStr .= "true"; }
						else
							{ $tmpStr .= "false"; }
						push @references, $tmpStr;
					}
				}
				if	($o->{pkey})
				{
					my	@opk = split /,/, $o->{pkey};
					@opk = map { strToPgId($_) } (@opk);
					print ",\n\tPRIMARY KEY(" . join(",", @opk) . ")";
				}
				print "\n)";

				if	(@{$o->{inherits}})
				{
					my	$y = 0;

					print " INHERITS (";
					foreach my $super (@{$o->{inherits}})
					{
						$y ? (print ", ") : ($y = 1);
						print strToPgId($super->{name});
					}
					print ")";
				}

				print ";\n";
				$tnam =~ s/\"//g;
				print "INSERT INTO d2p_lst_list(d2p_name, dia_name) VALUES ('$tnam', '$o->{name}');\n";
				foreach my $i (@{$o->{indices}})
				{
					print "CREATE " . ($i->{unique} ? 'UNIQUE ' : '') . "INDEX ";
					print strToPgId($i->{name}) . " ON " . strToPgId($o->{name});
					$i->{type} && print " USING " . $i->{type};
					print " (" . join(', ', map { strToPgId($_); } (@{$i->{fields}})) . ");\n";
				}
				foreach my $ref (@references)
				{
					$ref =~ s/\"//g;
					print "INSERT INTO d2p_ref_list(src_table, src_field, dest_table, dest_field, del_cascade, upd_cascade)\n";
					print "VALUES ($ref);\n";
				}
				print "\n";
			}
			print "\n";
		}
	}
}


###
## Create SQL footer
#
sub	copySQLFiles
{
	return unless @sqlfiles;

	$DEBUG && print STDERR " Copying additional SQL files\n";
	print "\n\n-- \n-- ADDITIONAL SQL FILES\n-- \n";
	foreach my $f (@sqlfiles)
	{
		unless	(open(SFIC, $f))
		{
			print STDERR "Warning : file $f not found ! \n";
			next;
		}
		print "\n-- \n -- SQL code from $f\n-- \n\n";
		my @cff;
		@cff = <SFIC>;
		close SFIC;
		print @cff;
	}
}


###
## Create SQL footer
#
sub	createSQLFooter
{
	$DEBUG && print STDERR " Writing footer\n";
	print "\n\n-- You may insert additionnal SQL code here (DO NOT REMOVE THIS LINE)\n";
	if	(@customCode)
	{
		print @customCode;
	}
	else
	{
		print "\n" x 4;
	}
	print "-- End of additionnal SQL code (DO NOT REMOVE THIS LINE)\n\n";
	print "SELECT f_d2p_inh_update();\nSELECT f_d2p_ref_update();\n\n";
	print "COMMIT TRANSACTION;\n";
	print "-- SQL code generated by dia2postgres.pl\n";
	print "-- $VERSION\n"; 
	print "-- on " . `date`;
}



###                  ###
#### PHP GENERATION ####
###                  ###

###
## Create complete PHP output
#
sub	createPhp
{
	$DEBUG && print STDERR "\nWriting PHP classes...\n";

	unless	(open(PHPOUT, ">$phpout"))
	{
		print STDERR "Warning : could not create PHP output file $phpout\n";
		return;
	}
	&createPHPHeader();
	&createPHPEnums();
	&createPHPClasses();
	&createPHPFooter();

	close PHPOUT;

	$DEBUG && print STDERR "Done!\n\n";
}


###
## Create PHP header
#
sub	createPHPHeader
{
	$DEBUG && print STDERR " Writing header\n";

	print PHPOUT "<?php\n\n/*\n";

	foreach my $g (@groups)
	{
		next unless ($g->{type} eq 'box' && $g->{topLevelText} ne '' && @{$g->{classes}} == 0);

		my	$title = $g->{topLevelText};
		$title =~ s/\n$//;

		my	@lines = split /\n/, $title;
		foreach my $l (@lines)
			{ print PHPOUT "$l\n"; }
		print PHPOUT "\n";
		last;
	}

	print PHPOUT "Input file  : $infile\n";
	print PHPOUT "\n";
	print PHPOUT "Defines PHP classes for the following modules : \n";
	foreach my $g (@groups)
	{
		next unless ($g->{type} eq 'box' && $g->{topLevelText} ne '' && @{$g->{classes}} != 0);

		my	$title = $g->{topLevelText};
		$title =~ s/\s+/ /g;

		print PHPOUT " - $title\n";
	}

	my	@requiredModules = ();
	foreach my $g (@groups)
	{
		next unless ($g->{type} eq 'pack' && $g->{topLevelText} ne '' && @{$g->{classes}} != 0);

		my	$title = $g->{topLevelText};
		$title =~ s/\s+/ /g;
		push @requiredModules, $title;
	}
	if	(@requiredModules)
	{
		print PHPOUT "\n";
		print PHPOUT "Requires the following modules :\n - ";
		print PHPOUT join ("\n - ", @requiredModules);
		print PHPOUT "\n";
	}
	print PHPOUT "*/\n\nrequire_once(\"d2p_manager.inc\");\n";
	print PHPOUT "require_once(\"d2p_data.inc\");";
	print PHPOUT "\n\n";
}


###
## Create PHP classes for enumerations according to data created by checkAndModifyInput()
#
sub	createPHPEnums
{
	foreach my $re (@enumerations)
	{
		my	$l = 0;
		my	$enum = {%$re};
		$enum->{tablename} =~ s/\"//g;
		foreach my $w1 (@{$enum->{contents}})
		{
			my	$w1l = length($w1);
			($w1l > $l) && ($l = $w1l);
		}

		print PHPOUT <<"ENDOFENUM";

// Enumeration class for member $enum->{fieldname}

class	d2p_cl_$enum->{tablename} extends d2p_enum_data
{
	function	d2p_cl_$enum->{tablename}(&\$manager, \$identifier = NULL)
	{
		\$this->d2p_data(\$manager, "$enum->{tablename}", "id");

		if      (!is_null(\$identifier))
		{
			\$db = \$this->manager->getDatabaseConnection();
			\$query = "SELECT * FROM \\"$enum->{tablename}\\" WHERE \\"id\\" = ";
			\$query .= \$this->fieldToSQL("int", \$identifier);
			\$result = \@pg_query(\$db, \$query);
			if      (!\$result)
			{
				\$this->errorMsg = "Unable to execute query : \$query";
				return  NULL;
			}
			\$row = pg_fetch_array(\$result);
			\$this->loadDatabaseEntry(\$row);
		}
	}
}
ENDOFENUM
	}
}


###
## Create PHP classes for each table
#
sub	createPHPClasses
{
	foreach my $gro (sort keys %groupsOrdered)
	{
		foreach my $g (@{$groupsOrdered{$gro}})
		{
			next unless ($g->{type} eq 'box' && @{$g->{classes}});

			my	$title = $g->{topLevelText};
			$title =~ s/\n/\n\/\/ /g;
			print PHPOUT "\n\n\n// Classes for module $title\n\n";
			foreach my $o (@{$g->{classes}})
			{
				my	$x = 0;
				my	@pkey = ();
				my	@codeMethodes = ();
				my	@codeChamps = ();
				my	@codeFormulaire = ();
				my	$tnam = &stripDQuotes(&strToPgId($o->{name}));

				print PHPOUT "class d2p_cl_$tnam extends ";
				if	(@{$o->{inherits}})
				{
					print PHPOUT "d2p_cl_" . &stripDQuotes(&strToPgId($o->{inherits}->[0]->{name}));
				}
				else
				{
					print PHPOUT "d2p_data";
				}
				print PHPOUT "\n{\n\tfunction\td2p_cl_$tnam(\&\$manager, \$identifier = NULL)\n\t{\n\t\t";
				print PHPOUT '$this->d2p_data($manager, ';
				{
					my	@opk = split /,/, $o->{pkey};
					@opk = map { &stripDQuotes(&strToPgId($_)) } (@opk);
					print PHPOUT '"' . $tnam . '", "' . $opk[0] . '");';
				}
				print PHPOUT <<"EOF";


		if      (!is_null(\$identifier))
		{
			\$db = \$this->manager->getDatabaseConnection();
			\$query = "SELECT * FROM \\"" . \$this->table . "\\" WHERE \\"" . \$this->identifierName . "\\" = ";
			\$query .= \$this->fieldToSQL(
				\$this->fields[\$this->identifierName]["type"], \$identifier
			);
			\$result = \@pg_query(\$db, \$query);
			if      (!\$result)
			{
				\$this->erreurFormat = "Unable to execute query : \$query";
				return  NULL;
			}
			\$row = pg_fetch_array(\$result);
			\$this->loadDatabaseEntry(\$row);
		}
	}

	function	prepareFields()
	{
		parent::prepareFields();
EOF

				foreach my $f (@{$o->{fields}})
				{
					my	$fnam = &stripDQuotes(&strToPgId($f->{name}));
					print PHPOUT "\t\t" . '$this->fields["' . $fnam . '"] = array(';
					print PHPOUT "\n\t\t\t\"initial_value\"\t=> NULL,";
					print PHPOUT "\n\t\t\t\"current_value\"\t=> NULL,";
					print PHPOUT "\n\t\t\t\"type\"\t\t=> \"";
					print PHPOUT ($f->{type} ne "serial" ? $f->{type} : "int");
					print PHPOUT "\",";

					if	($f->{references})
					{
						my	($table, $field) = map { &stripDQuotes(&strToPgId($_)); } (split (/\./, $f->{references}));
						print PHPOUT "\n\t\t\t\"references\"\t=> \"$table\",";
					}

					print PHPOUT "\n\t\t\t\"mandatory\"\t=> " . ($f->{not_null} ? "true" : "false") . "\n\t\t);\n";
					push @codeMethodes, "\n";
					push @codeMethodes, "\tfunction\tg_$fnam()\n";
					push @codeMethodes, "\t{\n";
					push @codeMethodes, "\t\treturn\t\$this->getField(\"$fnam\");\n";
					push @codeMethodes, "\t}\n";
					push @codeMethodes, "\n";
					push @codeMethodes, "\tfunction\ts_$fnam(\$value)\n";
					push @codeMethodes, "\t{\n";
					### FIXME : add type checks etc...
					push @codeMethodes, "\t\t\$this->setField(\"$fnam\", \$valeur);\n";
					push @codeMethodes, "\t}\n";

					push @codeFormulaire, "\t\t\$form .= \"<TR><TD>\\n\";\n";
					if	($f->{not_null})
					{
						push @codeFormulaire, "\t\t\$form .= \"<FONT COLOR='red'>\\n\";\n";
					}
					my $nomPresentable = $f->{name};
					$nomPresentable =~ s/_/ /g;
					push @codeFormulaire, "\t\t\$form .= \"$nomPresentable\\n\";\n";
					if	($f->{not_null})
					{
						push @codeFormulaire, "\t\t\$form .= \"</FONT>\\n\";\n";
					}
					push @codeFormulaire, "\t\t\$form .= \"</TD><TD>\\n\";\n";
					
					if	($f->{references})
					{
						my	($table, $field) = map { &stripDQuotes(&strToPgId($_)); } (split (/\./, $f->{references}));
						push @codeFormulaire, "\t\t\$form .= \"<SELECT NAME='form__$tnam\__$fnam'>\\n\";\n";
						push @codeFormulaire, "\t\t\$tab = \$this->manager->listFromBase"
							. "(\"$table\", \"SELECT \\\"$field\\\" FROM \\\"$table\\\"\");\n";
						push @codeFormulaire, "\t\tif\t(is_null(\$tab))\n";
						push @codeFormulaire, "\t\t\t\$form .= \"<OPTION VALUE=''>No records</OPTION>\\n\";\n";
						push @codeFormulaire, "\t\telse\n";
						push @codeFormulaire, "\t\t{\n";
						push @codeFormulaire, "\t\t\t\$form .= \"<OPTION VALUE=''>No record selected</OPTION>\\n\";\n";
						push @codeFormulaire, "\t\t\t\$o = \$this->getField(\"$fnam\");\n";
						push @codeFormulaire, "\t\t\tfor\t(\$i=0;\$i<count(\$tab);\$i++)\n";
						push @codeFormulaire, "\t\t\t{\n";
						push @codeFormulaire, "\t\t\t\t\$form .= \"<OPTION VALUE='\" . "
							. "\$tab[\$i]->getIdentifier() . \"'\";\n";
						push @codeFormulaire, "\t\t\t\tif\t(\$tab[\$i]->getObjId() == \$o->getObjId())\n";
						push @codeFormulaire, "\t\t\t\t\t\$form .= \" SELECTED\";\n";
						push @codeFormulaire, "\t\t\t\t\$form .= \">\" . "
							. "\$tab[\$i]->getDescription() . \"</OPTION>\\n\";\n";
						push @codeFormulaire, "\t\t\t}\n";
						push @codeFormulaire, "\t\t}\n";
						push @codeFormulaire, "\t\t\$form .= \"</SELECT>\\n\";\n";
					}
					else
					{
						push @codeFormulaire, "\t\t\$form .= \"<INPUT TYPE='text' NAME='form__$tnam";
						push @codeFormulaire, "__" . $fnam . "' VALUE=\\\"\";\n";
						push @codeFormulaire, "\t\t\$form .= \$this->getField(\"$fnam\");\n";
						push @codeFormulaire, "\t\t\$form .= \"\\\">\\n\";\n";
					}
					push @codeFormulaire, "\t\t\$form .= \"</TD></TR>\\n\";\n";
				}
				print PHPOUT "\t}\n";
				print PHPOUT join("", @codeMethodes);
				print PHPOUT "\n\tfunction\tgetHTMLForm()\n";
				print PHPOUT "\t{\n";
				print PHPOUT "\t\t\$form = parent::getHTMLForm();\n";
				print PHPOUT join("", @codeFormulaire);
				print PHPOUT "\t\treturn\t\$form;\n";
				print PHPOUT "\t}\n";
				print PHPOUT "}\n\n\n";
			}
		}
	}
}


###
## Create PHP footer
#
sub	createPHPFooter
{
	$DEBUG && print STDERR " Writing footer\n";
	print PHPOUT "\n\n// PHP code generated by dia2postgres.pl\n";
	print PHPOUT "// $VERSION\n"; 
	print PHPOUT "// on " . `date`;
	print PHPOUT "\n?>\n";
}



###                    ###
#### HELPER FUNCTIONS ####
###                    ###


###
## Parses command line arguments according to a description
#
sub	parseArguments
{
	my	$dptr	= shift @_;
	my	%desc	= %{$dptr};
	my	@args	= @_;
	my	$state	= 0;
	my	%opts	= ();
	my	%curarg	= ();
	my	$clel;
	my	$argc;

	return unless @args;

	$DEBUG && print STDERR "Command line arguments : '" . join("'    '", @args) . "'\n";

	foreach my $k (keys %desc)
	{
		my	@l = split /\|/, $k;
		foreach my $opt (@l)
		{
			$opts{$opt} = $k;
		}
	}

	foreach my $arg (@args)
	{
		$DEBUG && print STDERR "arg = $arg   state = $state\n";
		if	($state == 0)
		{
			if	(!grep(/^$arg$/, keys %opts))
			{
				die "Invalid argument : $arg\n";
			}

			my $k = $opts{$arg};
			%curarg = %{$desc{$k}};
			$DEBUG && print STDERR "curarg{uniq} = $curarg{uniq}    curarg{found} = $curarg{found}\n";
			die "Syntax error : $arg found twice in command line\n"
				if	($curarg{uniq} && $curarg{found});

			$desc{$k}->{found} = 1;

			if	($curarg{bvar})
			{
				my $cmd = "\$$curarg{bvar} = !$curarg{bvar};";
				eval $cmd;
			}
			elsif	($curarg{args} && !$curarg{argl})
			{
				$argc = 0;
				$state = 1;
			}
			elsif   ($curarg{args})
			{
				$argc = 0;
				$clel = {};
				$state = 2;
			}
			elsif	($curarg{hdlr})
			{
				&{$curarg{hdlr}} ($dptr);
			}
			else
			{
				print STDERR "Warning : $arg is not implemented\n";
			}
		}
		elsif	($state == 1)
		{
			# Standard argument
			die "Internal error in command line parsing\n"
				if	(!$curarg{args}->[$argc]);

			my $cmd = "\$" . $curarg{args}->[$argc] . " = \$arg;";
			eval "$cmd";

			$argc++;
			if ($argc == @{$curarg{args}})
			{
				&{$curarg{hdlr}} ($dptr) if $curarg{hdlr};
				$state = 0;
			}
		}
		elsif	($state == 2)
		{
			# List argument
			die "Internal error in command line parsing\n"
				if	(!$curarg{args}->[$argc]);

			my $cmd = "\$clel->{" . $curarg{args}->[$argc] . "} = \$arg;";
			eval $cmd;

			$argc++;
			if	($argc == @{$curarg{args}})
			{
				&{$curarg{hdlr}} ($dptr) if $curarg{hdlr};
				eval "push \@$curarg{argl}, \$clel;";
				$state = 0;
			}
		}
	}

	die "Error : unexpected end of command line\n"
		if	($state != 0);
}


###
## Shows command line help
#
sub		usage
{
	my	$dptr	= shift @_;
	my	%desc	= %{$dptr};
	my	$maxlen	= 0;
	my	$maxCol = $ENV{"COLUMNS"} ? $ENV{"COLUMNS"} : 80;

	$DEBUG && print STDERR "maxCol = $maxCol " . $ENV{COLUMNS} . "\n";

	print STDERR "dia2postgres - A Perl script that converts a Dia diagram into\n";
	print STDERR "               a PostgreSQL script, and maybe more.\n\n";
	print STDERR "$VERSION\n";
	print STDERR "(c) 2002 by Emmanuel Benoit <tseeker\@nocternity.net>\n\n";

	foreach my $a (keys %desc)
		{ $maxlen = length($a) if (length($a) > $maxlen); }
	$maxlen += 3;
	exit(1) if ($maxlen >= $maxCol - 15);

	print STDERR "Usage : $0 [<arguments>]\n\n";
	print STDERR "Where arguments are :\n";

	foreach my $arg (keys %desc)
	{
		my	@d	= split / /, $desc{$arg}->{desc};
		my	$col	= $maxlen + 1;
		my	$w;

		print STDERR " " . $arg . (" "x($maxlen - length($arg)));
		while	($w = shift @d)
		{
			if	($col + length($w) + 1 >= $maxCol)
			{
				$col = $maxlen + 4;
				print STDERR "\n" . (" "x$col);
			}
			print STDERR "$w ";
			$col += length($w) + 1;
		}
		print STDERR "\n";
	}
	print "\n\n";

	exit(1);
}


###
## Strip double quotes from a Postgres name
sub	stripDQuotes
{
	my $n = shift;
	$n =~ s/\"//g;
	return	$n;
}



###             ###
#### MAIN CODE ####
###             ###

###
## Command line arguments definitions and default values
#

%argsDescription = (
	"-i|--input"	=> {
		desc	=> "Read the Dia XML formatted input from the specified file instead of stdin.",
		args	=> ["infile"],
		uniq	=> 1
	},
	"-o|--input"	=> {
		desc	=> "Output the PostgreSQL script to the specified file instead of stdout.",
		args	=> ["outfile"],
		uniq	=> 1
	},
	"-s|--sql"	=> {
		desc	=> "Add the contents of the specified file at the end of the script. "
				. "May be repeated for each file to include to the script.",
		argl	=> "sqlfiles",
		args	=> ["sql"],
		uniq	=> 0
	},
	"-h|--help"	=> {
		desc	=> "Display this message.",
		uniq	=> 1,
		hdlr	=> \&usage
	},
	"-p|--php"	=> {
		desc	=> "Output PHP mirror classes to the specified file.",
		uniq	=> 1,
		args	=> ["phpout"]
	}
);

$infile = "";
$outfile = "";
$phpout = "";
@sqlfiles = ();

## Parse command line
&parseArguments(\%argsDescription, @ARGV);
@sqlfiles = map { $_->{sql} } @sqlfiles;

## Read input
@realin = ();
if	($infile ne "")
{
	open(FIC, $infile) or die "Error : could not open $infile\n";
	@realin = <FIC>;
	close(FIC);
}
else
{
	@realin = <STDIN>;
}

## Prepare output
if	($outfile)
{
	if	(-f "$outfile")
	{
		my	($s, @old);

		open(IN, "$outfile") or die "$outfile : $!\n";
		@old = <IN>;
		close(IN);

		$s = 0;
		foreach my $l (@old)
		{
			if	(!$s && ($l eq "-- You may insert additionnal SQL code here (DO NOT REMOVE THIS LINE)\n"))
			{
				$s = 1;
			}
			elsif	($s && ($l eq "-- End of additionnal SQL code (DO NOT REMOVE THIS LINE)\n"))
			{
				last;
			}
			elsif	($s)
			{
				push @customCode, $l;
			}
		}

		`/bin/mv -f "$outfile" "$outfile~"`;
	}
	open(STDOUT, ">$outfile") or die "$outfile: $!\n";
}

## Init xml parser and parse input file
$parser = new XML::Parser(Style => 'Stream');
$parser->parse(join("", @realin));
&checkAndModifyInput();

## Create sql output
&createSql();

## Create PHP code if required
&createPhp if ($phpout ne "");

## Cleanup and exit
$outfile && close(STDOUT);
exit;



